#!/usr/bin/env python

import logging
import os
from pprint import pformat


def set_log_config(debug=None, info=None):
    if debug:
        lv = logging.DEBUG
    elif info:
        lv = logging.INFO
    else:
        lv = logging.WARNING
    logging.basicConfig(
        format='%(asctime)s %(levelname)-8s %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S', level=lv
    )


def log_response(response, logger=None, expected_status_range=(100, 399)):
    logger = logger or logging.getLogger(__name__)
    res_str = 'response =>' + os.linesep + pformat(vars(response))
    esr = sorted(expected_status_range)
    if esr[0] <= response.status <= esr[-1]:
        logger.debug(res_str)
    else:
        logger.error(res_str)



#!/usr/bin/env python

import logging
import signal
import sqlite3
from abc import ABCMeta, abstractmethod
from datetime import datetime
from pathlib import Path

import pandas as pd
import redis
from v20 import V20ConnectionError, V20Timeout

from ..util.logger import log_response


class StreamDriver(object, metaclass=ABCMeta):
    def __init__(self, api, account_id, target='pricing', instruments=None,
                 timeout_sec=0, snapshot=True, ignore_api_error=False):
        if target not in ['pricing', 'transaction']:
            raise ValueError(f'invalid target:\t{target}')
        elif target == 'pricing' and not instruments:
            raise ValueError('pricing: instruments required')
        else:
            self.__logger = logging.getLogger(__name__)
            self.__api = api
            self.__account_id = account_id
            self.__target = target
            self.__instruments = instruments
            self.__timeout_sec = float(timeout_sec) if timeout_sec else None
            self.__snapshot = snapshot
            self.__ignore_api_error = ignore_api_error
            self.__latest_update_time = None

    def invoke(self):
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        try:
            res = self._call_stream_api()
            for msg_type, msg in res.parts():
                self.act(msg_type, msg)
                self.__latest_update_time = datetime.now()
        except (V20ConnectionError, V20Timeout) as e:
            if not self.__ignore_api_error:
                self.shutdown()
                raise e
            else:
                self.__logger.error(e)
                if self.__timeout_sec and self.__latest_update_time:
                    td = datetime.now() - self.__latest_update_time
                    if td.total_seconds() > self.__timeout_sec:
                        self.__logger.warning(
                            f'Timeout:\t{self.__timeout_sec} sec'
                        )
                        self.shutdown()
                        raise e
        else:
            log_response(res, logger=self.__logger)

    def _call_stream_api(self):
        if self.__target == 'pricing':
            self.__logger.info('Start to stream market prices')
            return self.__api.pricing.stream(
                accountID=self.__account_id, snapshot=self.__snapshot,
                instruments=','.join(self.__instruments)
            )
        elif self.__target == 'transaction':
            self.__logger.info('Start to stream transactions for the account')
            return self.__api.transaction.stream(accountID=self.__account_id)

    @abstractmethod
    def act(self, msg_type, msg):
        pass

    @abstractmethod
    def shutdown(self):
        pass


class StreamRecorder(StreamDriver):
    def __init__(self, api, account_id, target='pricing', instruments=None,
                 timeout_sec=0, snapshot=True, ignore_api_error=False,
                 skip_heartbeats=True, use_redis=False, redis_host='127.0.0.1',
                 redis_port=6379, redis_db=0, redis_max_llen=None,
                 sqlite_path=None, csv_path=None, quiet=False):
        super().__init__(
            api=api, account_id=account_id, target=target,
            instruments=instruments, timeout_sec=timeout_sec,
            snapshot=snapshot, ignore_api_error=ignore_api_error
        )
        self.__logger = logging.getLogger(__name__)
        self.__instruments = instruments
        self.__skip_heartbeats = skip_heartbeats
        self.__quiet = quiet
        if use_redis:
            self.__logger.info('Set a streamer with Redis')
            self.__redis_pool = redis.ConnectionPool(
                host=redis_host, port=int(redis_port), db=int(redis_db)
            )
            self.__redis_max_llen = (
                int(redis_max_llen) if redis_max_llen else None
            )
            redis_c = redis.StrictRedis(connection_pool=self.__redis_pool)
            redis_c.flushdb()
        else:
            self.__redis_pool = None
            self.__redis_max_llen = None
        if sqlite_path:
            self.__logger.info('Set a streamer with SQLite')
            sqlite_file = Path(sqlite_path).resolve()
            if sqlite_file.is_file():
                self.__sqlite = sqlite3.connect(str(sqlite_file))
            else:
                schema_sql = Path(__file__).parent.parent.joinpath(
                    'static/create_tables.sql'
                )
                self.__sqlite = sqlite3.connect(str(sqlite_file))
                with open(schema_sql, 'r') as f:
                    self.__sqlite.executescript(f.read())
        else:
            self.__sqlite = None
        if csv_path:
            self.__csv_path = str(Path(csv_path).resolve())
        else:
            self.__csv_path = None

    def act(self, msg_type, msg):
        if msg_type.endswith('Heartbeat') and self.__skip_heartbeats:
            self.__logger.debug(msg)
        elif (msg_type.startswith('transaction.') or
              (msg_type.startswith('pricing.') and msg.instrument)):
            self.__logger.debug(msg)
            self._print_and_write_msg(msg_type=msg_type, msg=msg)
        else:
            self.__logger.warning(f'Save skipped:\t{msg}')

    def _print_and_write_msg(self, msg_type, msg):
        msg_json_str = str(msg.json())
        if not self.__quiet:
            print(msg_json_str, flush=True)
        inst = (msg.instrument if hasattr(msg, 'instrument') else '')
        if self.__redis_pool:
            data_key = inst or 'transactions'
            redis_c = redis.StrictRedis(connection_pool=self.__redis_pool)
            redis_c.rpush(data_key, msg_json_str)
            if self.__redis_max_llen:
                if redis_c.llen(data_key) > self.__redis_max_llen:
                    redis_c.lpop(data_key)
        if self.__sqlite:
            c = self.__sqlite.cursor()
            table_name = msg_type.split('.')[0] + '_stream'
            c.execute(
                f'INSERT INTO {table_name} VALUES (?,?,?)',
                [msg.time, inst, msg_json_str]
            )
            self.__sqlite.commit()
        if self.__csv_path:
            pd.DataFrame(
                [{'time': msg.time, 'instrument': inst, 'json': msg_json_str}]
            ).set_index(
                ['time', 'instrument']
            ).to_csv(
                self.__csv_path, mode='a',
                sep=(',' if self.__csv_path.endswith('.csv') else '\t'),
                header=(not Path(self.__csv_path).is_file())
            )

    def shutdown(self):
        if self.__redis_pool:
            self.__redis_pool.disconnect()
        if self.__sqlite:
            self.__sqlite.close()


def invoke_streamer(api, account_id, instruments, target='pricing',
                    timeout_sec=0, csv_path=None, sqlite_path=None,
                    use_redis=False, redis_host='127.0.0.1', redis_port=6379,
                    redis_db=0, redis_max_llen=None, ignore_api_error=False,
                    quiet=False, skip_heartbeats=True):
    assert account_id, 'account ID required'
    assert instruments, 'instruments required'
    logger = logging.getLogger(__name__)
    logger.info('Streaming')
    if use_redis:
        assert redis_host, 'redis_host required'
        assert redis_port, 'redis_port required'
        assert redis_db or redis_db == 0, 'redis_db required'
    streamer = StreamRecorder(
        api=api, account_id=account_id, target=target, instruments=instruments,
        timeout_sec=timeout_sec, snapshot=True,
        ignore_api_error=ignore_api_error, skip_heartbeats=skip_heartbeats,
        use_redis=use_redis, redis_host=redis_host, redis_port=redis_port,
        redis_db=redis_db, redis_max_llen=redis_max_llen,
        sqlite_path=sqlite_path, csv_path=csv_path, quiet=quiet
    )
    streamer.invoke()



#!/usr/bin/env python

import json
import logging
import os
import sqlite3
import time
from pathlib import Path

import pandas as pd
import pandas.io.sql as pdsql
import yaml

from ..util.logger import log_response
from .plot import plot_pl


def track_transaction(api, account_id, from_time=None, to_time=None,
                      csv_path=None, sqlite_path=None, pl_graph_path=None,
                      print_json=False, quiet=False):
    assert account_id, 'account ID required'
    logger = logging.getLogger(__name__)
    logger.info('Transaction tracking')
    transactions = _fetch_transactions(
        api=api, account_id=account_id, from_time=from_time, to_time=to_time
    )
    if transactions:
        df_txn = pd.DataFrame([
            {'id': int(t['id']), 'time': t['time'], 'json': json.dumps(t)}
            for t in transactions
        ]).set_index('id')
        logger.debug(f'df_txn:{os.linesep}{df_txn}')
        if csv_path:
            if Path(csv_path).is_file():
                old_ids = set(
                    pd.read_csv(csv_path, usecols=['id'], dtype=int)['id']
                )
                df_txn.pipe(
                    lambda d: d[~d.index.isin(old_ids)]
                ).to_csv(csv_path, mode='a', header=False)
            else:
                df_txn.to_csv(csv_path)
        if sqlite_path:
            tbl = 'transaction_history'
            if Path(sqlite_path).is_file():
                with sqlite3.connect(sqlite_path) as con:
                    old_ids = set(
                        pdsql.read_sql(f'SELECT id FROM {tbl};', con=con)['id']
                    )
                    df_txn_new = df_txn.pipe(
                        lambda d: d[~d.index.isin(old_ids)]
                    )
                    logger.debug(f'df_txn_new:{os.linesep}{df_txn_new}')
                    if df_txn_new.size > 0:
                        pdsql.to_sql(
                            df_txn_new, name=tbl, con=con, if_exists='append'
                        )
            else:
                schema_sql = Path(__file__).parent.parent.joinpath(
                    'static/create_tables.sql'
                )
                with sqlite3.connect(sqlite_path) as con:
                    with open(schema_sql, 'r') as f:
                        con.executescript(f.read())
                    pdsql.to_sql(df_txn, name=tbl, con=con, if_exists='append')
        if pl_graph_path:
            plot_pl(df_txn=df_txn, path=pl_graph_path)
    if not quiet:
        print(
            json.dumps(transactions, indent=2) if print_json
            else yaml.dump(transactions, default_flow_style=False).strip()
        )


def _fetch_transactions(api, account_id, from_time=None, to_time=None):
    logger = logging.getLogger(__name__)
    res = api.transaction.list(
        accountID=account_id,
        **{
            k: v for k, v
            in {'fromTime': from_time, 'toTime': to_time}.items() if v
        }
    )
    log_response(res, logger=logger)
    transactions = list()
    for page in (res.body.get('pages') or list()):
        time.sleep(0.5)
        r = api.transaction.range(
            accountID=account_id, **_parse_idrange(page=page)
        )
        log_response(r, logger=logger)
        transactions.extend(
            json.loads(r.raw_body).get('transactions') or list()
        )
    return transactions


def _parse_idrange(page):
    return dict([
        s.split('=') for s in page.split('?')[1].replace('=', 'ID=').split('&')
    ])



#!/usr/bin/env python

import json
import logging
import os
import sqlite3
from itertools import product

import matplotlib.pyplot as plt
import pandas as pd
import pandas.io.sql as pdsql
import seaborn as sns
from matplotlib.lines import Line2D


def read_and_plot_pl(data_path, graph_path):
    exts = {
        'csv': tuple([f'.csv{s}' for s in ['', '.gz', '.bz2']]),
        'tsv': tuple([
            (a + b) for a, b in product(['.tsv', '.txt'], ['', '.gz', '.bz2'])
        ]),
        'sqlite': ('.sqlite3', '.sqlite', '.db')
    }
    if data_path.endswith(exts['csv']):
        df_txn = pd.read_csv(
            data_path, sep=',', usecols=['json'], dtype={'json': str}
        )
    elif data_path.endswith(exts['tsv']):
        df_txn = pd.read_csv(
            data_path, sep='\t', usecols=['json'], dtype={'json': str}
        )
    elif data_path.endswith(exts['sqlite']):
        with sqlite3.connect(data_path) as con:
            df_txn = pdsql.read_sql(
                'SELECT json FROM transaction_history;', con=con
            )
    else:
        raise ValueError(f'unsupported file type:\t{data_path}')
    plot_pl(df_txn=df_txn, path=graph_path)


def plot_pl(df_txn, path):
    logger = logging.getLogger(__name__)
    df_pl = pd.DataFrame([
        {
            **{
                k: o.get(k)
                for k in ['time', 'instrument', 'accountBalance', 'pl']
            },
            'initialMarginRequired': (
                o['tradeOpened'].get('initialMarginRequired')
                if o.get('tradeOpened') else None
            )
        } for o in df_txn['json'].apply(json.loads)
    ]).pipe(
        lambda d: d[d['accountBalance'].notna() & d['instrument'].notna()]
    ).astype(
        dtype={
            k: float for k in ['accountBalance', 'pl', 'initialMarginRequired']
        }
    ).assign(
        time=lambda d: pd.to_datetime(d['time'])
    )
    logger.debug(f'df_pl:{os.linesep}{df_pl}')
    df_cumpl = df_pl.set_index(['instrument', 'time'])['pl'].unstack(
        level=0, fill_value=0
    ).cumsum().stack().to_frame('pl').reset_index()
    logger.debug(f'df_cumpl:{os.linesep}{df_cumpl}')

    plt.rcParams['figure.figsize'] = (11.88, 8.40)  # A4 aspect: (297x210)
    sns.set(style='ticks', color_codes=True)
    sns.set_context('paper')
    fig, axes = plt.subplots(nrows=3)
    plt.subplots_adjust(hspace=0.6)
    instruments = set(df_pl['instrument'])
    colors = {
        k: v for k, v in zip(
            sorted(instruments),
            sns.color_palette('deep', n_colors=len(instruments)).as_hex()
        )
    }
    alpha = 0.7
    time_range = (df_pl['time'].min(), df_pl['time'].max())
    ylim_ratio = 1.2

    for i, d in df_cumpl.groupby('instrument'):
        axes[0].plot(
            'time', 'pl', label=i, color=colors[i], data=d, alpha=alpha
        )
    axes[0].set(
        title='Cumulative Profit and Loss', ylabel='pl', xlim=time_range,
        ylim=(pd.Series([-1, 1]) * df_cumpl['pl'].abs().max() * ylim_ratio)
    )

    for i, d in df_pl.groupby('instrument'):
        axes[1].scatter(
            x='time', y='initialMarginRequired', label=i, color=colors[i],
            data=d, alpha=alpha, marker='+'
        )
    axes[1].set(
        title='Initial Margins of Trades',
        ylabel='initialMarginRequired', xlim=time_range,
        ylim=(0, df_pl['initialMarginRequired'].max() * ylim_ratio)
    )

    axes[2].fill_between(
        x='time', y1='accountBalance', color='lightsteelblue', data=df_pl
    )
    axes[2].set(
        title='Account Balance', xlabel='time', ylabel='accountBalance',
        xlim=time_range, ylim=(0, df_pl['accountBalance'].max() * ylim_ratio)
    )

    sns.despine()
    fig.legend(
        handles=[
            Line2D([0], [0], color=v, label=k, alpha=alpha)
            for k, v in colors.items()
        ],
        loc='upper left', bbox_to_anchor=(0.91, 0.91), ncol=1,
        title='instrument'
    )
    fig.savefig(path, bbox_inches='tight')



#!/usr/bin/env python

import logging
import os
from pprint import pformat

from ..util.logger import log_response


def close_positions(api, account_id, instruments=None):
    assert account_id, 'account ID required'
    logger = logging.getLogger(__name__)
    logger.info('Position closing')
    pos_res = api.position.list_open(accountID=account_id)
    log_response(pos_res, logger=logger)
    positions = pos_res.body['positions']
    insts_to_close = list()
    if positions:
        for p in positions:
            if not instruments or p.instrument in instruments:
                insts_to_close.append(p.instrument)
                pos = {
                    'instrument': p.instrument,
                    **{
                        f'{k}Units':
                        ('NONE' if int(getattr(p, k).units) == 0 else 'ALL')
                        for k in ['long', 'short']
                    }
                }
                logger.debug(f'pos:\t{pos}')
                res = api.position.close(accountID=account_id, **pos)
                log_response(res, logger=logger)
                if 100 <= res.status <= 399:
                    logger.debug(res.body)
                else:
                    raise RuntimeError(
                        'unexpected response:' + os.linesep + pformat(res.body)
                    )
    if insts_to_close:
        print('Positions closed:\t' + ', '.join(insts_to_close))
    else:
        print('No positions to close.')



#!/usr/bin/env python
"""
Command Line Interface for Oanda API

Usage:
    oanda-cli -h|--help
    oanda-cli --version
    oanda-cli init [--debug|--info] [--file=<yaml>]
    oanda-cli info [--debug|--info] [--file=<yaml>] [--json] <info_target>
                   [<instrument>...]
    oanda-cli track [--debug|--info] [--file=<yaml>] [--csv-dir=<path>]
                    [--sqlite=<path>] [--granularity=<code>] [--count=<int>]
                    [--json] [--quiet] [<instrument>...]
    oanda-cli stream [--debug|--info] [--file=<yaml>] [--target=<str>]
                     [--timeout=<sec>] [--csv=<path>] [--sqlite=<path>]
                     [--use-redis] [--redis-host=<ip>] [--redis-port=<int>]
                     [--redis-db=<int>] [--redis-max-llen=<int>]
                     [--ignore-api-error] [--quiet] [<instrument>...]
    oanda-cli transaction [--debug|--info] [--file=<yaml>] [--from=<date>]
                          [--to=<date>] [--csv=<path>] [--sqlite=<path>]
                          [--pl-graph=<path>] [--json] [--quiet]
    oanda-cli plotpl [--debug|--info] <data_path> <graph_path>
    oanda-cli spread [--debug|--info] [--file=<yaml>] [--csv=<path>] [--quiet]
                     [<instrument>...]
    oanda-cli close [--debug|--info] [--file=<yaml>] [<instrument>...]

Options:
    -h, --help          Print help and exit
    --version           Print version and exit
    --debug, --info     Execute a command with debug|info messages
    --file=<yaml>       Set a path to a YAML for configurations [$OANDA_YML]
    --quiet             Suppress messages
    --csv-dir=<path>    Write data with daily CSV in a directory
    --sqlite=<path>     Save data in an SQLite3 database
    --granularity=<code>
                        Set a granularity for rate tracking [default: S5]
    --count=<int>       Set a size for rate tracking (max: 5000) [default: 60]
    --json              Print data with JSON
    --target=<str>      Set a streaming target [default: pricing]
                        { pricing, transaction }
    --timeout=<sec>     Set senconds for response timeout
    --csv=<path>        Write data with CSV into a file
    --use-redis         Use Redis for data store
    --redis-host=<ip>   Set a Redis server host (override YAML configurations)
    --redis-port=<int>  Set a Redis server port (override YAML configurations)
    --redis-db=<int>    Set a Redis database (override YAML configurations)
    --redis-max-llen=<int>
                        Limit Redis list length (override YAML configurations)
    --ignore-api-error  Ignore Oanda API connection errors
    --from=<date>       Specify the starting time
    --to=<date>         Specify the ending time
    --pl-graph=<path>   Visualize PL in a graphics file such as PDF or PNG

Commands:
    init                Create a YAML template for configuration
    info                Print information about <info_target>
    track               Fetch past rates
    stream              Stream market prices or authorized account events
    transaction         Fetch the latest transactions
    plotpl              Visualize cumulative PL in a file
    spread              Print the ratios of spread to price
    close               Close positions (if not <instrument>, close all)

Arguments:
    <info_target>       { instruments, prices, account, accounts, orders,
                          trades, positions, position, order_book,
                          position_book }
    <instrument>        { AUD_CAD, AUD_CHF, AUD_HKD, AUD_JPY, AUD_NZD, AUD_SGD,
                          AUD_USD, CAD_CHF, CAD_HKD, CAD_JPY, CAD_SGD, CHF_HKD,
                          CHF_JPY, CHF_ZAR, EUR_AUD, EUR_CAD, EUR_CHF, EUR_CZK,
                          EUR_DKK, EUR_GBP, EUR_HKD, EUR_HUF, EUR_JPY, EUR_NOK,
                          EUR_NZD, EUR_PLN, EUR_SEK, EUR_SGD, EUR_TRY, EUR_USD,
                          EUR_ZAR, GBP_AUD, GBP_CAD, GBP_CHF, GBP_HKD, GBP_JPY,
                          GBP_NZD, GBP_PLN, GBP_SGD, GBP_USD, GBP_ZAR, HKD_JPY,
                          NZD_CAD, NZD_CHF, NZD_HKD, NZD_JPY, NZD_SGD, NZD_USD,
                          SGD_CHF, SGD_HKD, SGD_JPY, TRY_JPY, USD_CAD, USD_CHF,
                          USD_CNH, USD_CZK, USD_DKK, USD_HKD, USD_HUF, USD_INR,
                          USD_JPY, USD_MXN, USD_NOK, USD_PLN, USD_SAR, USD_SEK,
                          USD_SGD, USD_THB, USD_TRY, USD_ZAR, ZAR_JPY }
    <data_path>         Path to an input CSV or SQLite file
    <graph_path>        Path to an output graphics file such as PDF or PNG
"""

import logging
import os
from pathlib import Path

import v20
from docopt import docopt

from .. import __version__
from ..call.candle import track_rate
from ..call.info import print_info, print_spread_ratios
from ..call.order import close_positions
from ..call.plot import read_and_plot_pl
from ..call.streamer import invoke_streamer
from ..call.transaction import track_transaction
from ..util.config import fetch_config_yml_path, read_yml, write_config_yml
from ..util.logger import set_log_config


def main():
    args = docopt(__doc__, version=f'oandacli {__version__}')
    set_log_config(debug=args['--debug'], info=args['--info'])
    logger = logging.getLogger(__name__)
    logger.debug(f'args:{os.linesep}{args}')
    config_yml_path = fetch_config_yml_path(path=args['--file'])
    execute_command(args=args, config_yml_path=config_yml_path)


def execute_command(args, config_yml_path):
    if args.get('init'):
        write_config_yml(
            dest_path=config_yml_path,
            template_path=str(
                Path(__file__).parent.parent.joinpath(
                    'static/default_oanda.yml'
                )
            )
        )
    else:
        config = read_yml(path=config_yml_path)
        api = v20.Context(
            hostname='{0}-fx{1}.oanda.com'.format(
                ('stream' if args.get('stream') else 'api'),
                config['oanda']['environment']
            ),
            token=config['oanda']['token']
        )
        account_id = config['oanda'].get('account_id')
        instruments = (
            args.get('<instrument>') or config.get('instruments') or list()
        )
        if args.get('info'):
            print_info(
                api=api, account_id=account_id, instruments=instruments,
                target=args['<info_target>'], print_json=args['--json']
            )
        elif args.get('spread'):
            print_spread_ratios(
                api=api, account_id=account_id, instruments=instruments,
                csv_path=args['--csv'], quiet=args['--quiet']
            )
        elif args.get('track'):
            track_rate(
                api=api, instruments=instruments,
                granularity=args['--granularity'], count=args['--count'],
                csv_dir_path=args['--csv-dir'], sqlite_path=args['--sqlite'],
                print_json=args['--json'], quiet=args['--quiet']
            )
        elif args.get('stream'):
            rd = config.get('redis') or dict()
            invoke_streamer(
                api=api, account_id=account_id, instruments=instruments,
                target=args['--target'], timeout_sec=args['--timeout'],
                csv_path=args['--csv'], sqlite_path=args['--sqlite'],
                use_redis=args['--use-redis'],
                redis_host=(args['--redis-host'] or rd.get('host')),
                redis_port=(args['--redis-port'] or rd.get('port')),
                redis_db=(args['--redis-db'] or rd.get('db')),
                redis_max_llen=args['--redis-max-llen'],
                ignore_api_error=args['--ignore-api-error'],
                quiet=args['--quiet']
            )
        elif args.get('transaction'):
            track_transaction(
                api=api, account_id=account_id, from_time=args['--from'],
                to_time=args['--to'], csv_path=args['--csv'],
                sqlite_path=args['--sqlite'], pl_graph_path=args['--pl-graph'],
                print_json=args['--json'], quiet=args['--quiet']
            )
        elif args.get('plotpl'):
            read_and_plot_pl(
                data_path=args['<data_path>'], graph_path=args['<graph_path>']
            )
        elif args.get('close'):
            close_positions(
                api=api, account_id=account_id, instruments=instruments
            )



#!/usr/bin/env python

import logging
import os
import shutil
from pathlib import Path

import yaml


def read_yml(path):
    with open(path, 'r') as f:
        d = yaml.load(f, Loader=yaml.FullLoader)
    return d


def write_config_yml(dest_path, template_path):
    logger = logging.getLogger(__name__)
    if Path(dest_path).exists():
        print(f'A file already exists:\t{dest_path}')
    else:
        logger.info(f'Write a config:\t{dest_path}')
        shutil.copyfile(template_path, dest_path)
        print(f'A YAML template was generated:\t{dest_path}')


def fetch_config_yml_path(path=None, env='OANDA_YML', default='oanda.yml'):
    logger = logging.getLogger(__name__)
    p = [
        str(Path(p).resolve()) for p in [path, os.getenv(env), default]
        if p is not None
    ][0]
    logger.debug(f'abspath to a config:\t{p}')
    return p



#!/usr/bin/env python

import json
import logging

import pandas as pd
import yaml

from ..util.logger import log_response


def print_info(api, account_id=None, instruments=None, target='accounts',
               print_json=False):
    logger = logging.getLogger(__name__)
    available_targets = [
        'instruments', 'account', 'accounts', 'orders', 'trades', 'positions',
        'transactions', 'prices', 'position', 'order_book', 'position_book'
    ]
    if target not in available_targets:
        raise ValueError(f'invalid info target:\t{target}')
    logger.info('Information')
    arg_insts = {'instruments': ','.join(instruments)} if instruments else {}
    logger.debug(f'information target:\t{target}')
    if target == 'instruments':
        res = api.account.instruments(accountID=account_id, **arg_insts)
    elif target == 'account':
        res = api.account.get(accountID=account_id)
    elif target == 'accounts':
        res = api.account.list()
    elif target == 'orders':
        res = api.order.list_pending(accountID=account_id)
    elif target == 'trades':
        res = api.trade.list_open(accountID=account_id)
    elif target == 'positions':
        res = api.position.list_open(accountID=account_id)
    elif not instruments:
        raise ValueError(f'{target}:\tinstruments required')
    elif target == 'prices':
        res = api.pricing.get(accountID=account_id, **arg_insts)
    elif target == 'position':
        res = api.position.get(accountID=account_id, instrument=instruments[0])
    elif target == 'order_book':
        res = api.instrument.order_book(instrument=instruments[0])
    elif target == 'position_book':
        res = api.instrument.position_book(instrument=instruments[0])
    log_response(res, logger=logger)
    data = json.loads(res.raw_body)
    print(
        json.dumps(data, indent=2) if print_json
        else yaml.dump(data, default_flow_style=False).strip()
    )


def print_spread_ratios(api, account_id, instruments=None, csv_path=None,
                        quiet=False):
    assert account_id, 'account ID required'
    logger = logging.getLogger(__name__)
    logger.info('Prices and Spread Ratios')
    if instruments:
        insts = instruments
    else:
        res0 = api.account.instruments(accountID=account_id)
        log_response(res0, logger=logger)
        insts = [o['name'] for o in json.loads(res0.raw_body)['instruments']]
    res1 = api.pricing.get(accountID=account_id, instruments=','.join(insts))
    log_response(res1, logger=logger)
    df_spr = pd.DataFrame([
        {k: o[k] for k in ['instrument', 'closeoutBid', 'closeoutAsk']}
        for o in json.loads(res1.raw_body)['prices']
    ]).rename(
        columns={'closeoutBid': 'bid', 'closeoutAsk': 'ask'}
    ).astype(
        dtype={'bid': float, 'ask': float}
    ).assign(
        mid=lambda d: d[['bid', 'ask']].mean(axis=1),
        spread=lambda d: (d['ask'] - d['bid'])
    ).assign(
        ratio_of_spread_to_mid=lambda d: (d['spread'] / d['mid'])
    ).set_index('instrument').sort_values('ratio_of_spread_to_mid')
    if csv_path:
        df_spr.to_csv(csv_path)
    if not quiet:
        with pd.option_context('display.max_rows', None):
            print(df_spr)



#!/usr/bin/env python

import json
import logging
import os
import sqlite3
from itertools import chain
from pathlib import Path

import pandas as pd
import pandas.io.sql as pdsql

from ..util.logger import log_response


def track_rate(api, instruments, granularity, count, csv_dir_path=None,
               sqlite_path=None, print_json=False, quiet=False):
    assert instruments, 'instruments required'
    logger = logging.getLogger(__name__)
    logger.info('Rate tracking')
    candles = dict()
    for i in instruments:
        res = api.instrument.candles(
            instrument=i, price='BA', granularity=granularity, count=int(count)
        )
        log_response(res, logger=logger)
        candles[i] = [
            _candlestick2dict(c) for c in res.body['candles'] if c.complete
        ]
    keys = ['instrument', 'time']
    df_all = pd.concat([
        pd.DataFrame(c).assign(instrument=i) for i, c in candles.items()
    ]).drop(columns=['complete']).set_index(keys)
    if csv_dir_path:
        csv_dir = Path(csv_dir_path).resolve()
        if not csv_dir.is_dir():
            csv_dir.mkdir()
        df_all_day = df_all.reset_index().assign(
            datetime=lambda d: pd.to_datetime(d['time'])
        ).assign(
            date=lambda d: d['datetime'].dt.date
        )
        for t in df_all_day['date'].unique():
            for i in df_all_day['instrument'].unique():
                df_csv = df_all_day.pipe(
                    lambda d: d[(d['date'] == t) & (d['instrument'] == i)]
                ).drop(columns=['date', 'instrument'])
                csv_path = str(
                    csv_dir.joinpath(f'candle.{granularity}.{i}.{t}.csv')
                )
                if Path(csv_path).is_file():
                    df_csv_new = df_csv.append(
                        pd.read_csv(csv_path).assign(
                            datetime=lambda d: pd.to_datetime(d['time'])
                        )
                    ).sort_values('datetime').drop(
                        columns='datetime'
                    ).drop_duplicates(
                        subset=['time'], keep='last'
                    ).set_index('time')
                else:
                    df_csv_new = df_csv.sort_values('datetime').drop(
                        columns='datetime'
                    ).set_index('time')
                df_csv_new.to_csv(csv_path, mode='w', header=True, sep=',')
    if sqlite_path:
        logger.debug(f'df_all.shape:\t{df_all.shape}')
        sqlite_file = Path(sqlite_path).resolve()
        if sqlite_file.is_file():
            with sqlite3.connect(str(sqlite_file)) as con:
                df_db_diff = df_all.join(
                    pdsql.read_sql(
                        'SELECT instrument, time FROM candle;', con
                    ).assign(
                        in_db=True
                    ).set_index(keys),
                    on=keys, how='left'
                ).pipe(
                    lambda d: d[d['in_db'].isna()].drop(columns=['in_db'])
                )
                logger.debug(f'df_db_diff:{os.linesep}{df_db_diff}')
                pdsql.to_sql(
                    df_db_diff, name='candle', con=con, if_exists='append'
                )
        else:
            schema_sql = Path(__file__).parent.parent.joinpath(
                'static/create_tables.sql'
            )
            with sqlite3.connect(str(sqlite_file)) as con:
                with open(schema_sql, 'r') as f:
                    con.executescript(f.read())
                logger.debug(f'df_all:{os.linesep}{df_all}')
                pdsql.to_sql(
                    df_all, name='candle', con=con, if_exists='append'
                )
    if not quiet:
        if print_json:
            print(json.dumps(candles, indent=2))
        else:
            with pd.option_context('display.max_rows', None):
                print(df_all)


def _candlestick2dict(candlestick):
    data_keys = ['bid', 'ask', 'mid']
    abbr = {'o': 'open', 'h': 'high', 'l': 'low', 'c': 'close'}
    cs_dict = candlestick.dict()
    return {
        **{k: v for k, v in cs_dict.items() if k not in data_keys},
        **dict(
            chain.from_iterable([
                [(abbr[k] + dk.capitalize(), float(v)) for k, v in d.items()]
                for dk, d in cs_dict.items() if dk in data_keys and d
            ])
        )
    }
